#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  4 16:39:05 2019

@author: ben
"""

import os
import pickle
import traceback

FILE_PATH = "files/"

def init():
    if not os.path.isdir("./" + FILE_PATH):
        os.mkdir("./" + FILE_PATH)
    
def save(grid, file):
    if not file.endswith('.vw'):
        file = file + ".vw"
    try: 
        with open(FILE_PATH + file, 'wb') as f:
             pickle.dump(grid, f)
    except:
        traceback.print_exc()
    
def load(file):
    if not file.endswith('.vw'):
        file = file + ".vw"
    try:
        with open(FILE_PATH + file, 'rb') as f:
            return pickle.load(f)
    except:
        traceback.print_exc()


def exists(file):
    file = format_file(file)
    return file in files()
    
def format_file(file):
    if not file.endswith('.vw'):
        file = file + ".vw"
    return file

def files():
    try:
        f = [file for file in os.listdir('./' + FILE_PATH) if file.endswith(".vw")]
        f.sort()
        return f
    except:
        traceback.print_exc()
        return []




# -*- coding: utf-8 -*-
import sys, os
myPath = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, myPath + '/../')


from ..vw import Grid
from ..vwc import coord


def test_wall_locations_exist():
    grid = Grid(5)
    assert coord(-1,-1) in grid.state.keys() or coord(-1, 5) in grid.state.keys() or coord(5,-1) in grid.state.keys() or  coord(5, 5) in grid.state.keys()
    
def test_location_exist(): # every location is properly assigned to a coordinate within the grid
    grid = Grid(5)
    assert not grid.state[coord(2, 2)]==None 

def test_agent_added_properly(): # whenever agent method is called agent is added properly to the grid
    grid = Grid(5)
    grid.agent('green','north')
    assert grid.agent_count==1 
   # assert grid.agent_count==2 
def test_agentPlacedProperly(): # whenever agent is placed Is  it placed properly
    grid = Grid(5)
    agent1 = grid.agent('green','north')
    grid.replace_agent((coord(2,1)), agent1)
    loc = grid.state[coord(2,1)]
    assert loc.agent==agent1
   # assert loc.agent==None  # in case of not properly placed agent
 
def test_dirtaddedProperly(): # whenever agent method is called agent is added properly to the grid
    grid = Grid(5)
    grid.dirt('green')
   # grid.dirt('white')     # should not be added
    assert grid.dirt_count==1 
   # assert grid.dirt_count==2      

def test_dirtPlacedProperly(): # whenever agent is placed Is  it placed properly
    grid = Grid(5)
    dirt1 = grid.dirt('green')
    grid.replace_dirt((coord(2,1)), dirt1)
    loc = grid.state[coord(2,1)]
    assert loc.dirt==dirt1
   # assert loc.agent==None  # in case of not properly placed agent
   
def test_dirtRemovedProperly():
    grid = Grid(4)
    dirt1 = grid.dirt('orange')
    grid.replace_dirt((coord(1,1)), dirt1)
    grid.remove_dirt(coord(1,1))
    loc = grid.state[coord(1,1)]
    assert loc.dirt==None

def test_agentRemovedProperly():
    grid = Grid(4)
    agent1 = grid.agent('orange','north')
    grid.replace_agent((coord(1,1)), agent1)
    grid.remove_agent(coord(1,1))
    loc = grid.state[coord(1,1)]
    assert loc.agent==None
def test_agentMoveProperly():
    grid = Grid(4)
    agent1 = grid.agent('orange','east')
    grid.replace_agent((coord(0,0)), agent1)
    grid.move_agent((coord(0,0)),(coord(1,0)))
    assert grid.state[coord(0,0)].agent==None   # location vacated
    assert grid.state[coord(1,0)].agent!=None   # location occupied 
    
def test_agentTurnProperly():
    grid = Grid(4)
    agent1 = grid.agent('orange','east')
    grid.replace_agent((coord(0,0)), agent1)
    grid.turn_agent((coord(0,0)),'south')
    assert grid.state[coord(0,0)].agent.orientation=='south'   # orientation has changed 
   # assert grid.state[coord(0,0)].agent.orientation=='south'  # orientation is same changed     
        
           
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jun  2 23:16:51 2019

@author: ben
"""


from .vwc import location, dirt, agent, coord
from .vwc import colour as vwcolour
from .vwc import orientation as vworientation


import random
from inspect import signature
from .vwenvironment import GridEnvironment, GridAmbient

#--------------------------------------------------------

AGENT_COLOURS = set(['user', 'orange', 'green', 'white'])
DIRT_COLOURS = set(['orange', 'green'])

def init(grid, minds):
    return GridEnvironment(GridAmbient(grid, minds))

def minds(mind):
    return {colour:mind for colour in AGENT_COLOURS}

class VacuumWorldInternalError(Exception):
    pass
    
#TODO throw errors instead of returning
def __validate_mind(agent, colour):
    
    def decide_def(fun):
        if not callable(fun):
            raise VacuumWorldInternalError("{0} agent: decide must be callable".format(colour))            
        if len(signature(fun).parameters) != 0:
            raise VacuumWorldInternalError("{0} agent: decide must be defined with no arguments, do(self)".format(colour))
    
    def revise_def(fun):
        if not callable(fun):
            raise VacuumWorldInternalError("{0} agent: revise must be callable".format(colour, fun.__name__))            
        if len(signature(fun).parameters) != 2:
            raise VacuumWorldInternalError("{0} agent: revise must be defined with two arguments, revise(self, observation, messages)".format(colour))
             
    MUST_BE_DEFINED = {'decide':decide_def, 
                       'revise':revise_def}
    agent_dir = set(dir(agent))

    for fun,validate in MUST_BE_DEFINED.items():
        if fun in agent_dir:
            validate(getattr(agent, fun))
        else:
            raise VacuumWorldInternalError("{0} agent: must define method: {1}".format(colour, fun))
        
    return True    

def random_grid(size, green, white, orange, user, orange_dirt, green_dirt):
    assert green + white + orange + user <= size
    assert orange_dirt + green_dirt <= size
    grid = Grid(size)
    #for agents
    coords = random.choices([key for key,value in grid.state.items() if value is not None] , k = green + white + orange)
    agents = {vwcolour.white:white, vwcolour.orange:orange, 
              vwcolour.green:green, vwcolour.user:user}
    for c, num in agents.items():
        for j in range(num):
            grid.place_agent(coords.pop(-1), grid.agent(c, random.choice(vworientation)))
    #for dirts
    dirts = {vwcolour.orange:orange_dirt,vwcolour.green:green}
    coords = random.choices([key for key,value in grid.state.items() if value is not None] , k = orange_dirt + green_dirt)
    for colour, num in dirts.items():
        for j in range(num):
            grid.place_dirt(coords.pop(-1), grid.dirt(colour))
    return grid

class Grid:
    DIRECTIONS = {'north':(0,-1), 'south':(0,1), 'west':(-1,0), 'east':(1,0)}
    
    ID_PREFIX_DIRT = 'D-'
    ID_PREFIX_AGENT = 'A-'
    
    GRID_MIN_SIZE = 3
    GRID_MAX_SIZE = 13
    
    def __init__(self, dim):
        self.state = None
        self.reset(dim)
        self.agent_count = 0
        self.dirt_count = 0 
        self.cycle = 0
       
    def replace_all(self, grid):
        self.dim = grid.dim
        self.state = grid.state
        self.agent_count = grid.agent_count
        self.dirt_count = grid.dirt_count
        self.cycle = grid.cycle
       
    def reset(self, dim):
        self.cycle = 0
        self.state = {}
        for i in range(dim):
            for j in range(dim):
                self.state[coord(j,i)] = location(coord(j,i), None, None)
            self.state[coord(-1, i)] = None
            self.state[coord(i, -1)] = None
            self.state[coord(dim, i)] = None
            self.state[coord(i, dim)] = None
        self.state[coord(-1,-1)] = None
        self.state[coord(-1, dim)] = None
        self.state[coord(dim,-1)] = None
        self.state[coord(dim, dim)] = None
        self.dim = dim
        self.agent_count = 0
        self.dirt_count = 0
        
    def _get_agents(self):
        return {coord:location.agent for coord,location in self.state.items() if location and location.agent}
    
    def _get_dirts(self):
        return {coord:location.dirt for coord,location in self.state.items() if location and location.dirt}
        
    def _in_bounds(self, coordinate):
        return coordinate.x >= 0 and coordinate.x < self.dim and coordinate.y >= 0 and coordinate.y < self.dim
    
    def _as_coord(self, coordinate):
        if not isinstance(coordinate, coord):
            return coord(coordinate[0], coordinate[1])
        return coordinate
    
    def dirt(self, colour):
        assert(colour in DIRT_COLOURS)
        self.dirt_count += 1
        return dirt(Grid.ID_PREFIX_DIRT + str(self.dirt_count), colour)
    
    def agent(self, colour, direction):
        assert(colour in AGENT_COLOURS)
        assert(direction in Grid.DIRECTIONS.keys())
        self.agent_count += 1
        return agent(Grid.ID_PREFIX_AGENT + str(self.agent_count), colour, direction)            
    
    def replace_agent(self, coordinate, agent):
         coordinate = self._as_coord(coordinate)
         assert(self._in_bounds(coordinate))
         loc = self.state[coordinate]
         self.state[coordinate] = location(coordinate, agent, loc.dirt)
         
    def replace_dirt(self, coordinate, dirt):
         coordinate = self._as_coord(coordinate)
         assert(self._in_bounds(coordinate))
         loc = self.state[coordinate]
         self.state[coordinate] = location(coordinate, loc.agent, dirt)
        
    def place_agent(self, coordinate, agent):
        coordinate = self._as_coord(coordinate)
        assert(self._in_bounds(coordinate))
        assert(self.state[coordinate].agent == None)
        loc = self.state[coordinate]
        self.state[coordinate] = location(coordinate, agent, loc.dirt)
        
    def place_dirt(self, coordinate, dirt):
        coordinate = self._as_coord(coordinate)
        assert(self._in_bounds(coordinate))
        assert(self.state[coordinate].dirt == None)
        loc = self.state[coordinate]
        self.state[coordinate] = location(coordinate, loc.agent, dirt)
    
    def remove_dirt(self, coordinate):
        assert(self._in_bounds(coordinate))
        loc = self.state[coordinate]
        self.state[coordinate] = location(coordinate, loc.agent, None)
        
    def remove_agent(self, coordinate):
        assert(self._in_bounds(coordinate))
        loc = self.state[coordinate]
        self.state[coordinate] = location(coordinate, None, loc.dirt)
        
    def move_agent(self, _from, _to):
        _from = self._as_coord(_from)
        _to = self._as_coord(_to)
        assert(self.state[_from].agent != None)
        assert(self.state[_to].agent == None)
        
        from_loc = self.state[_from]
        to_loc = self.state[_to]
        self.state[_to] = location(to_loc.coordinate, from_loc.agent, to_loc.dirt)
        self.state[_from] = location(from_loc.coordinate, None, from_loc.dirt)
        
    def turn_agent(self, _coordinate, orientation):
        assert(self.state[_coordinate].agent != None)
        loc = self.state[_coordinate]
        ag = loc.agent
        self.state[_coordinate] = location(_coordinate, agent(ag.name, ag.colour, orientation), loc.dirt)
    



    def __str__(self):
        header = "{0}: size: {1}, agents: {2}, dirts: {3}, ".format(str(type(self)), self.dim, self.agent_count, self.dirt_count)
        filled = {coord:location for coord,location in self.state.items()  if location is not None and (location.agent is not None or location.dirt is not None)}
        body = "\n".join([str(location) for location in filled.values()])
        return "\n".join([header, body])
        
    def __repr__(self):
        return str(self)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jun  2 21:23:17 2019

@author: Benedict Wilkins
"""

from . import vw
from . import vwv
from . import vwc
from . import vwutils

__all__ = ('vw', 'vwv', 'vwc')

__MARKING = False
__CURRENT = None
__OBSERVE = {'grid_size':None}


def run(white_mind, green_mind=None, orange_mind=None, **kwargs):

    def __validate_mind(mind, colour, observe):
        for obs in observe:
            if not obs in set(dir(mind)):
                raise vw.VacuumWorldInternalError("{0} agent: must define the attribute: {1}, (see coursework question 1)".format(colour, obs)) 
        def sneaky_setattr(self, name, value):
            if name in observe: 
                if observe[name] == None:
                    print("Agent {0} updated {1}: {2}".format(vwutils.callerID(), name, value))
                else:
                    observe[name](mind, colour, name, value)
            super(type(mind), self).__setattr__(name, value)
            
        type(mind).__setattr__ = sneaky_setattr
        
        return vw.__validate_mind(mind, colour)
        
    if green_mind is None:
        green_mind = white_mind
    if orange_mind is None:
        orange_mind = white_mind
    
    if __MARKING:
        __MARKING_HASH = r'c0bd538fd8bdaac7a98802fd9f0ddda29aa6bea3705e88b9dc32ab5d4821fa58'
        __MARKING_PASSWORD_FILE = 'password.txt'
        import hashlib
        with open(__MARKING_PASSWORD_FILE, 'rb') as f:
             password = f.readline()
             m = hashlib.sha256(password)
             if m.hexdigest() == __MARKING_HASH:
                 #import vacuumworldmarking #import the marking module
                 global __CURRENT
                 if __CURRENT is not None:
                      raise ValueError("Student called run more than once... oh dear")
                 __CURRENT = (white_mind, green_mind, orange_mind)  
             else:
                 raise ValueError('Incorrect marking password')
    else:
        white_ok = __validate_mind(white_mind, 'white', __OBSERVE)    
        green_ok = __validate_mind(green_mind, 'green', __OBSERVE)
        orange_ok = __validate_mind(orange_mind, 'orange', __OBSERVE)
        
        #must contain grid_size... for the coursework, really we should place this somewhere else as it is not general!
        
        if white_ok and green_ok and orange_ok:
            vwv.run({vwc.colour.white:white_mind, 
                     vwc.colour.green:green_mind, 
                     vwc.colour.orange:orange_mind}, **kwargs)







from warnings import warn

from pystarworlds.Event import Action, Executor
from . import vwc
from . import vwutils

###################### action executors ###################### 

class MoveExecutor(Executor):
    
    def __call__(self, env, action):
        agent = env.ambient.agents[action.__actor__]
        
        new_coordinate = agent.coordinate + vwc.orientation_map[agent.orientation]
        new_location = env.ambient.grid.state[new_coordinate]
        #precondition
        if new_location and not new_location.agent:
            env.ambient.grid.move_agent(agent.coordinate, new_coordinate)
            agent.coordinate = new_coordinate
        
class TurnExecutor(Executor):
         
    def __call__(self, env, action):
        agent = env.ambient.agents[action.__actor__]
        new_orientation = action.direction(agent.orientation)
        env.ambient.grid.turn_agent(agent.coordinate, new_orientation)
        agent.orientation = new_orientation

class CleanExecutor(Executor):
        
    def __call__(self, env, action):
        agent = env.ambient.agents[action.__actor__]
        #precondition
        location = env.ambient.grid.state[agent.coordinate]
        if location.dirt and (location.agent.colour == location.dirt.colour or location.agent.colour == vwc.colour.white):  
            env.ambient.grid.remove_dirt(agent.coordinate)
            #TODO remove dirt from list of objects in ambient

class DropExecutor(Executor):
        
    def __call__(self, env, action):
        agent = env.ambient.agents[action.__actor__]
        #precondition
        if not env.ambient.grid.state[agent.coordinate].dirt: 

            env.ambient.grid.place_dirt(agent.coordinate, env.ambient.grid.dirt(action.colour))
            #location = env.ambient.grid.state[agent.coordinate]
           
            #TODO add new dirt to list of objects in ambient
        
class CommunicativeExecutor(Executor):

    def __call__(self, env, action):
        notify = action.to
        if len(notify) == 0: #send to everyone! do we want this?
            notify = set(env.ambient.agents.keys()) - set([action.__actor__])
        for to in notify:
            env.physics.notify_agent(env.ambient.agents[to], vwc.message(action.__actor__, action.content))
            
###################### actions ###################### 
            
class VWPhysicalAction(Action):
     def __init__(self):
         super(VWPhysicalAction, self).__init__()
    
class MoveAction(VWPhysicalAction):
    executor = MoveExecutor
    pass
             
class CleanAction(VWPhysicalAction):
    executor = CleanExecutor
    pass

class DropAction(VWPhysicalAction):
    executor = DropExecutor
    
    def __init__(self, colour):
        super(DropAction, self).__init__()
        assert colour in vwc._colour_dirt
        self.colour = colour
        
class TurnAction(VWPhysicalAction):
    executor = TurnExecutor
    def __init__(self, direction):
        super(TurnAction, self).__init__()
        assert direction in vwc.direction
        self.direction = direction

class CommunicativeAction(Action):
    executor = CommunicativeExecutor
    
    def __init__(self, content, *to):
        super(CommunicativeAction, self).__init__()
        self.content = content
        self.to = to

###################### create actions from label ###################### 

class ActionFactory:
    
    def __init__(self):
        pass
    
    def __call__(self):
        pass
    
class DropActionFactory(ActionFactory):
    
    def __init__(self):
        super(DropActionFactory, self).__init__()
        
    def __call__(self, _colour):
        return DropAction(_colour)
    
class TurnActionFactory(ActionFactory):
    
    def __init__(self):
        super(TurnActionFactory, self).__init__()
        
    def __call__(self, _direction):
        return TurnAction(_direction)
    
class SpeakActionFactory(ActionFactory):
    
    LIMIT = 100
    
    def __init__(self):
        super(SpeakActionFactory, self).__init__()
        
    def __call__(self, _message, *_to):
        self.__validate_message(_message)
        _size = vwc.size(_message)
        if _size > SpeakActionFactory.LIMIT:
            _osize = _size
            _message, _size = self.__chop(_message)
            vwutils.warn_agent("WARNING: Agent: {0} Message trimmed from {1} to {2}", _osize, _size)
        for t in _to:
            assert(isinstance(t, str))
        return CommunicativeAction(_message, *_to)
    
    def __validate_message(self, message):
        assert(type(message) in (str, int, float, bool, list, tuple, type(None)))
        if type(message) in (list, tuple, set):
            for e in message:
               self.__validate_message(e)
               
    def __chop(self, message):
        _size = vwc.size(message)

        while _size > SpeakActionFactory.LIMIT and type(message) in (list, tuple):
            if len(message) == 1:
                message = message[0]
                break
            message, _size = self.__chop(message[:-1])
        
        if vwc.size(message) > SpeakActionFactory.LIMIT:
            _t = type(message)
            message = _t(str(message)[:SpeakActionFactory.LIMIT])
            _size = vwc.size(message)
        return message, _size

_action_factories = {vwc.move.__name__:lambda: MoveAction(), 
                     vwc.clean.__name__:lambda: CleanAction(), 
                     vwc.idle.__name__:lambda: None,
                     vwc.turn.__name__:TurnActionFactory(), 
                     vwc.drop.__name__:DropActionFactory(),
                     vwc.speak.__name__:SpeakActionFactory()}



    
    


# -*- coding: utf-8 -*-
from pystarworlds.Perception import Perception

#maybe we use this... who knows!
class ActionResultPerception(Perception):
    
    def __init__(self,agid,action,result):
        super().__init__(agid)
        self.__actionAttempted__=action
        self.__result__=result   
     
    def getResult(self):
        return self.__result__
    
    def getActionAttempted(self):
        return self.__actionAttempted__
    


from pystarworlds.Agent import Sensor

from .vwc import observation, message

class VisionSensor(Sensor):
    subscribe = [observation]
    pass

class CommunicationSensor(Sensor):  # ear for getting a communication perception
    subscribe = [message]
    pass


from pystarworlds.Agent import Mind, Body

from . import vwsensor
from . import vwactuator
from . import vwagent 
from . import vwaction
from . import vwc

from collections import namedtuple
agent_type = namedtuple('agent_type', 'cleaning user')('cleaning', 'user')

class VWBody(Body):
    
    def __init__(self, _type, ID, mind, orientation, coordinate, colour):
        mind = vwagent.VWMind(mind)
        assert _type in agent_type
        if _type == agent_type.cleaning:
            actuators = {"physical":vwactuator.CleaningAgentActuator(),
                         "communication":vwactuator.CommunicationActuator()}
        else:
            actuators = {"physical":vwactuator.UserActuator(),
                         "communication":vwactuator.CommunicationActuator()}

        sensors = {"vision":vwsensor.VisionSensor(),
                   "communication":vwsensor.CommunicationSensor()}
        self.ID = ID
        
        super(VWBody, self).__init__(mind, actuators, sensors)
        
        self.orientation = orientation
        self.colour = colour 
        self.coordinate = coordinate

class VWMind(Mind):
    
    def __init__(self, surrogate, observers = []):
        super(VWMind, self).__init__()
        self.surrogate = surrogate
        self.observers = []
        for observer in observers:
            self.observers.append()
       
    def cycle(self):
 
        observation = [percept for percept in self.body._sensors["vision"]]
        assert(len(observation) == 1)
        
        messages = [percept for percept in self.body._sensors["communication"]]
        self.surrogate.revise(*observation, messages)
    
        actions = self.surrogate.decide()

        if actions is None:
            return
        
        speak_actions = []
        physical_actions = []

        if isinstance(actions, tuple):
            if type(actions) in vwc.action_types:
               actions = [actions]
            for action in actions:
                if action is None:
                    continue
                if type(action) in vwc.action_types:
                    if type(action) == vwc.idle:
                        continue
                    if type(action) == vwc.speak:
                        _a = vwaction._action_factories[type(action).__name__](action.message, *action.to)
                        speak_actions.append((_a, type(action).__name__))
                    else:
                        _a = vwaction._action_factories[type(action).__name__](*action[:])
                        physical_actions.append((_a, type(action).__name__))
                elif callable(action):
                    raise ActionError(str(action) + " should not be callable, did you forget ()?")
                else:
                    raise ActionError(str(action) + " is an invalid action.")
                    
            _assert(len(speak_actions) <= 1, ' an agent can perform at most one speech action per cycle, attempted:' + str([s[1] for s in speak_actions]))
            _assert(len(physical_actions) <= 1, 'an agent can perform at most one physical action per cycle, attempted:' +  str([s[1] for s in physical_actions]))
            
            for action in speak_actions + physical_actions:
                actuators = self.body.find_actuators(action[0])
                _assert(len(actuators) == 1,  str(action[1]) + ' for agent - actuator does not exist')
                actuators[0].attempt(action[0])
        else:
            raise ActionError(str(actions) + " is an invalid action.")
            
            
        
        
        
def validate_action(action):
    pass
    #_assert(not callable(action), 'invalid action, did you forget to call the action? - e.g. move()')
    #_assert(isinstance(action, tuple), str(action) + 'please make use of vwc.action')
   #_assert(action[0] in vwc.action_names,  str(action) + 'must be one of ' + str(list(vwc.action_names)) + ' from vwc.action')

class ActionError(Exception):
    
    def __init__(self, message):
        super(ActionError, self).__init__("Invalid action: " + message)

def _assert(condition, exception):
    if not condition:
        raise ActionError(exception)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jun  1 20:05:16 2019

@author: ben
"""
import tkinter as tk

class Slider(tk.Canvas):
    
    def __init__(self, parent, release_callback, slide_callback, img, width, height, increments = 0, slider_width = 8, start = 0, **kwargs):
        super(Slider, self).__init__(parent, width=width, height=height, bd=0, highlightthickness=0, relief='ridge', bg='white', **kwargs)
        self.release_callback = release_callback
        self.slide_callback = slide_callback
        
        self.increments = increments
        self.slider_image = img
        self.slide_item_dim = slider_width
        
        self.x = start #real position of the slider
        self.inc = 0   #incremental position of the slider [0-increments] 
        if start > width:
            start = width - self.slide_item_dim/2
        if increments: 
            dx = int((width - self.slide_item_dim) / self.increments)
            self.x = start * dx
            self.inc = start
            

        self.background_item = self.create_rectangle(0,0, width-1, height-1, fill='white')
        self.slider_item = self.create_rectangle(self.x, 0, self.x + self.slide_item_dim, height, fill='black') #self.create_image(img, 0, 0)
        self.bind("<ButtonPress-1>", self.on_start)
        self.bind("<B1-Motion>", self.on_drag)
        self.bind("<ButtonRelease-1>", self.on_drop)
    
    def set_position(self, inc, callback=True):
        if inc != self.inc:
            inc = max(0, min(inc, self.increments))
            width = self.winfo_width() - self.slide_item_dim
            old_x = self.x
            self.x = inc * int(width / self.increments)
            self.inc = inc
            self.move(self.slider_item, self.x - old_x, 0)
            if callback:
                self.slide_callback(self.inc)
                self.release_callback(self.inc)
        
    def _move_slider(self, event):
        width = self.winfo_width() - self.slide_item_dim
        x = event.x     
        if x > 0 and x < width:
            if self.increments:
                inc = int(width / self.increments)
                self.inc = int(x / inc) 
                x = self.inc * inc

            dx = x - self.x
            if dx != 0:
                self.move(self.slider_item, dx, 0)
                self.slide_callback(self.inc)
                self.x = x
    
    def on_start(self, event):
        self._move_slider(event)
        
    def on_drag(self, event):
        self._move_slider(event)
        
    def on_drop(self, event):
        x = self.x
        if self.increments:
            x = self.inc
        self.release_callback(int(x))
       



from . import vwaction as action
from . import vwagent
from . import vwc
from . import vwuser

from pystarworlds.Environment import Ambient, Physics, Environment, Process
from pystarworlds.Identifiable import Identifiable

import copy


#from VWFactories import VWGridPerceptionFactory,ForwardActionRuleFactory,TurnLeftActionRuleFactory,TurnRightActionRuleFactory,CleanActionRuleFactory,DropActionRuleFactory,SpeakActionRuleFactory,SpeakToAllActionRuleFactory,ForwardActionExecuteFactory,TurnLeftActionExecuteFactory,TurnRightActionExecuteFactory,CleanActionExecuteFactory,DropActionExecuteFactory,SpeakActionExecuteFactory,SpeakToAllActionExecuteFactory
#from GridPerception import Observation,Message,ActionResultPerception
#from GridWorldAction import ForwardMoveMentAction,MoveLeftAction,MoveRightAction, CleanDirtAction,DropDirtAction, SpeakAction,BroadcastAction
#from vw import Direction
#from pystarworlds.Identifiable import Identifiable

def init(grid, minds, user_mind):
    minds[vwc.colour.user] = vwuser.USERS[user_mind]()
    return GridEnvironment(GridAmbient(grid, minds))
    
class GridAmbient(Ambient):
    
    def __init__(self, grid, minds):
        self.grid = grid
        agents=[]
        dirts=[]
        for location in grid.state.values(): 
            if location:
                if location.agent:       
                    agents.append(self.init_agent(location, self.get_type(location), minds))
                if location.dirt:       
                    dirts.append(Dirt(location.dirt))
        super(GridAmbient, self).__init__(agents, dirts)
    
    def get_type(self, location):
        return vwagent.agent_type[int(location.agent.colour == vwc.colour.user)]
        
    def init_agent(self, location, _type, minds):
        return vwagent.VWBody(_type, location.agent.name, copy.deepcopy(minds[location.agent.colour]),
                              location.agent.orientation, location.coordinate, location.agent.colour)
        

class GridPhysics(Physics):
    pass

class GridEnvironment(Environment):
    
    def __init__(self, ambient):
        
        actions = [action.DropAction, action.MoveAction,
                   action.TurnAction, action.CleanAction, 
                   action.CommunicativeAction]
        
        physics = GridPhysics(actions)
        
        super(GridEnvironment, self).__init__(physics, ambient, [ObservationProcess()], actions)

class Dirt(Identifiable):
    
    def __init__(self, dirt):
        self.ID = dirt.name
        super(Dirt, self).__init__()
        self.dirt = dirt

class ObservationProcess(Process):
    
    def __call__(self, env):
        for agent in env.ambient.agents.values():
            env.physics.notify_agent(agent, self.__get_perception__(env.ambient, agent))
            
    def __get_perception__(self, ambient, agent):
        c = agent.coordinate
        f = vwc.orientation_map[agent.orientation]
        l = vwc.orientation_map[vwc.left(agent.orientation)]
        r = vwc.orientation_map[vwc.right(agent.orientation)]
        #center left right forward forwardleft forwardright
        obs = vwc.observation(ambient.grid.state[c], 
                        ambient.grid.state[c + l],
                        ambient.grid.state[c + r], 
                        ambient.grid.state[c + f], 
                        ambient.grid.state[c + f + l],
                        ambient.grid.state[c + f + r])            
        return obs





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
CREDIT: http://code.activestate.com/recipes/578253-an-entry-with-autocompletion-for-the-tkinter-gui/
"""

import tkinter as tk
import re

class AutocompleteEntry(tk.Entry):
    def __init__(self, lista, height, *args, **kwargs):
        tk.Entry.__init__(self, *args, **kwargs)
        self.dropdown_parent = args[0]
        self.lista = lista        
        self.var = self["textvariable"]
        if self.var == '':
            self.var = self["textvariable"] = tk.StringVar()
        self.var.trace('w', self.changed)
        self.height = height
        #self.bind("<Key>", self.changed)
        self.bind("<Right>", self.selection)
        self.bind("<Return>", self.selection)
        self.bind("<Up>", self.up)
        self.bind("<Down>", self.down)
        self.lb = None
        self.lb_up = False

    def changed(self, *args):
        #print(args)
        #name, index, mode = args
        #print("changed", name, index, mode, self.var.get())

        if self.var.get() == '':
            if self.lb:
                self.lb.destroy()
            self.lb_up = False
        else:
            words = self.comparison()
            if words:            
                if not self.lb_up:
                    self.lb = tk.Listbox(self.dropdown_parent, height=self.height)
                    
                    self.lb.bind("<Double-Button-1>", self.selection)
                    print(self.winfo_x(), self.winfo_y()+self.winfo_height())
                    self.lb.place(x=self.winfo_x(), y=self.winfo_y()+self.winfo_height())
                    #self.lb.place(x=381, y=300)
                    self.lb_up = True
                
                self.lb.delete(0, tk.END)
                for w in words:
                    self.lb.insert(tk.END,w)
            else:
                if self.lb_up:
                    self.lb.destroy()
                    self.lb_up = False
        
    def selection(self, event):
        #print(event)
        if self.lb_up:
            self.var.set(self.lb.get(tk.ACTIVE))
            self.lb.destroy()
            self.lb_up = False
            self.icursor(tk.END)

    def up(self, event):
        if self.lb_up and not self.lb.curselection() == ():
            #print('up', self.lb.curselection())
            index = self.lb.curselection()[0]
            if index >= 0:
                self.lb.yview_scroll(-1, "units")
                self.lb.selection_clear(first=index)
                index = str(int(index)-1)                
                self.lb.selection_set(first=index)
                self.lb.activate(index) 


    def down(self, event):
        if self.lb_up:
            #print('down', self.lb.curselection(), len(self.lista))
            if self.lb.curselection() == ():
                self.lb.selection_set(first=0)
                self.lb.activate(0) 
                return

            index = self.lb.curselection()[0]
            if int(index) + 1 < len(self.lista):
                if int(index) + 1 >= self.height:
                    self.lb.yview_scroll(1, "units") 
                    #print('scroll')
                
                self.lb.selection_clear(first=index)
                index = str(int(index)+1)        
                self.lb.selection_set(first=index)
                self.lb.activate(index) 
                
    def comparison(self):
        #print("compare")
        pattern = re.compile('.*' + self.var.get() + '.*')
        return [w for w in self.lista if re.match(pattern, w)]

if __name__ == '__main__':
    root = tk.Tk()
    lista = ['a', 'actions', 'additional', 'also', 'an', 'and', 'angle', 'are', 'as', 'be', 'bind', 
             'bracket', 'brackets', 'button', 'can', 'cases', 'configure', 'course', 'detail', 
             'enter', 'event', 'events', 'example', 'field', 'fields', 'for', 'give', 'important', 
             'in', 'information', 'is', 'it', 'just', 'key', 'keyboard', 'kind', 'leave', 'left', 
             'like', 'manager', 'many', 'match', 'modifier', 'most', 'of', 'or', 'others', 'out', 
             'part', 'simplify', 'space', 'specifier', 'specifies', 'string;', 'that', 'the',
             'there', 'to', 'type', 'unless', 'use', 'used', 'user', 'various', 'ways', 'we', 'window', 'wish', 'you']

    entry = AutocompleteEntry(lista, root)
    entry.grid(row=0, column=0)
    tk.Button(text='nothing').grid(row=1, column=0)
    tk.Button(text='nothing').grid(row=2, column=0)
    tk.Button(text='nothing').grid(row=3, column=0)

    root.mainloop()





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 31 20:12:24 2019

@author Benedict Wilkins
"""

import tkinter as tk
import traceback
import os
import time
from threading import Thread, Event

from collections import OrderedDict as odict
from PIL import Image, ImageTk


#from Slider import Slider
#from vw import Environment
from .slider import Slider
from .autocomplete import AutocompleteEntry
from .vw import Grid
from .vwenvironment import init as init_environment
from . import vwc
from . import saveload
from . import vwuser

#might need to change this for the real package...
PATH = os.path.dirname(__file__)
TIME_STEP_MIN = 0.04
TIME_STEP_BASE = 1. - TIME_STEP_MIN #in seconds
TIME_STEP = TIME_STEP_BASE
TIME_STEP_MODIFIER = 1.

WIDTH = 640
HEIGHT = 480
ROOT_FONT = "Verdana 10 bold" #font.Font(family='Helvetica', size=36, weight='bold')
BUTTON_PATH = PATH + "/res/"
LOCATION_AGENT_IMAGES_PATH = PATH + "/res/locations/agent"
LOCATION_DIRT_IMAGES_PATH = PATH + "/res/locations/dirt"
MAIN_MENU_IMAGE_PATH = PATH + "/res/start_menu.png"
DEFAULT_LOCATION_SIZE = 60
DEFAULT_GRID_SIZE = 480
BACKGROUND_COLOUR_SIDE = 'white'
BACKGROUND_COLOUR_GRID = 'white'

DIFFICULTY_LEVELS = len(vwuser.USERS)
INITIAL_ENVIRONMENT_DIM = 8

def get_location_img_files(path):
    return [file for file in os.listdir(path) if file.endswith(".png")]

class VWButton:

    def __init__(self, root, img, fun, text = None):
        self.img = img
        self.fun = fun
        self._button = tk.Button(root,
                                 text = text,
                                 bd=0,
                                 font = ROOT_FONT,
                                 fg='white',
                                 highlightthickness = 0,
                                 bg='white',
                                 activebackground='white',
                                 activeforeground='white',
                                 highlightcolor='white',
                                 compound = 'center',
                                 command = fun)
        self._button.config(image=img)

    def pack(self, side):
        self._button.pack(side=side)
        
    def grid(self, row, col):
        self._button.grid(row=row,column=col)
        
    def destroy(self):
        self._button.destroy()
        self.img.destroy()

class VWDifficultyButton(VWButton):

    def __init__(self, root, img, fun):
        self.imgs = [ImageTk.PhotoImage(img)]
        self.imgs.extend([self.next_image(img, i * (255/(DIFFICULTY_LEVELS-1))) for i in range(1, DIFFICULTY_LEVELS)])
        super(VWDifficultyButton, self).__init__(root, self.imgs[0], self.onclick)
        self.difficulty = 0
        self._rfun = fun

    def next_image(self, img, red):
        img_dif = Image.new("RGB", img.size)
        img_dif.paste(img)
        rr, gg, bb = img_dif.split()
        rr = rr.point(lambda p: p + red)
        img_dif = Image.merge("RGB", (rr, gg, bb))
        return ImageTk.PhotoImage(img_dif)

    def onclick(self):
        self.difficulty = (self.difficulty + 1) % len(self.imgs)
        self.img = self.imgs[self.difficulty]
        self._button.config(image=self.img)
        self._rfun() #update the global variable .... TODO change this


class VWMainMenu(tk.Frame):

    def __init__(self, root, _start, _exit):
        super(VWMainMenu,self).__init__(root)
        self.configure(background='white')
        WIDTH = 480
        HEIGHT = 480
        self.canvas = tk.Canvas(self, width=WIDTH + 1,
                                height=HEIGHT + 1,
                                bd=0,
                                highlightthickness=0)
        #self.canvas.create_rectangle(0,0,481,481,fill="blue") # placeholder for image
        
        self.image_tk = tk.PhotoImage(file=MAIN_MENU_IMAGE_PATH)
        self.image = self.canvas.create_image(WIDTH/2,HEIGHT/2,image=self.image_tk)
        
        self.button_frame = tk.Frame(self)
        
        #self.canvas.create_window(WIDTH/2, HEIGHT-100, window =  self.button_frame)
        self.canvas.pack()
        self.buttons = {}

        button_image = ImageTk.PhotoImage(Image.open(BUTTON_PATH + 'button.png'))
        
        self.buttons['start'] = VWButton(self.button_frame, button_image, _start, 'start')
        self.buttons['exit'] = VWButton(self.button_frame, button_image, _exit, 'exit')
        self.buttons['start'].pack('left')
        self.buttons['exit'].pack('left')
        self.button_frame.pack()

        self.pack()


class CanvasDragManager:

    def __init__(self, key, grid, canvas, item, on_start, on_drop):
        self.x = 0
        self.y = 0
        self.canvas = canvas

        self._on_start = on_start
        self._on_drop = on_drop
        self.canvas.tag_bind(item, "<ButtonPress-1>", self.on_start)
        self.canvas.tag_bind(item, "<B1-Motion>", self.on_drag)
        self.canvas.tag_bind(item, "<ButtonRelease-1>", self.on_drop)
        #self.canvas.configure(cursor="hand1")

        self.key = key
        self.drag_image = None
        self.drag = None
        self.dragging = False
        self.grid = grid

    def on_start(self, event):
        if not self.dragging:
            self._on_start(event)
            self.dragging = True
            self.x = event.x
            self.y = event.y

    def on_drag(self, event):

        inc = DEFAULT_GRID_SIZE / self.grid.dim
        x = int(event.x / inc) * inc + (inc / 2) + 1
        y = int(event.y / inc) * inc + (inc / 2) + 1

        if x != self.x or y != self.y:
            if x <= DEFAULT_GRID_SIZE:
                self.canvas.itemconfigure(self.drag, state='normal')
            else:
                self.canvas.itemconfigure(self.drag, state='hidden')
            dx = x - self.x
            dy = y - self.y
            self.canvas.move(self.drag, dx, dy)
            self.x = x
            self.y = y


    def on_drop(self, event):
        #print('drop')
        if _in_bounds(event.x, event.y):
            self._on_drop(event, self)
        self.dragging = False


class VWInterface(tk.Frame):

    SIDE_PANEL_WIDTH = DEFAULT_LOCATION_SIZE + 4

    def __init__(self, parent, grid):
        super(VWInterface, self).__init__(parent)
        self.parent = parent
        self.configure(background=BACKGROUND_COLOUR_SIDE)
        self.canvas = tk.Canvas(self, width=DEFAULT_GRID_SIZE+VWInterface.SIDE_PANEL_WIDTH,
                                height=DEFAULT_GRID_SIZE + 1,
                                bd=0,
                                highlightthickness=0)
        #self.canvas.create_rectangle(0,0,481,481,fill="blue") # placeholder for grid
        #self.drag = self.canvas.create_rectangle(230,230,10,10, fill='yellow')

        self._init_buttons()

        self.canvas.configure(background=BACKGROUND_COLOUR_GRID)

        self.grid = grid

        self.canvas_dirts = {}
        self.canvas_agents = {}

        self.all_images = {} #stores all PIL images
        self.all_images_tk = {} #stores all tk images
        self.all_images_tk_scaled = {} #stores all tk images scale to fit grid
        self.grid_lines = [] #stores line objects

        self._init_images()
        self._init_dragables()

        self._draw_grid(grid.dim)

        self.canvas.grid(row=0,column=0) #packing

        #bind keys for rotation
        parent.bind('<Left>', self.rotate_agent_left)
        parent.bind('<Right>', self.rotate_agent_right)
        parent.bind('<a>', self.rotate_agent_left)
        parent.bind('<d>', self.rotate_agent_right)
        self.parent = parent

        self.canvas.bind('<Double-Button-1>', self.remove_top)
        self.canvas.bind('<Button-1>', self.select)
        self.canvas.bind('<Motion>', self.on_mouse_move)
        self.canvas.bind("<Leave>", self.on_leave_canvas)
        
        
        self.currently_selected = None
        self.running = False
        self.rectangle_selected = None
        self.selected = None
        
    def _init_buttons(self):
        self.buttons = {}

        bg = 'white'
        buttons = get_location_img_files(BUTTON_PATH)
        buttons = {b.split('.')[0]:b for b in buttons}

        self.button_frame = tk.Frame(self, bg=bg)


        play_tk = tk.PhotoImage(file=BUTTON_PATH + buttons['play']) # ImageTk.PhotoImage(Image.open(BUTTON_PATH + buttons['play'])) #
        #play_tk = Image.open(BUTTON_PATH + buttons['play'])
        
        #left side contains buttons and slider
        self.left_frame = tk.Frame(self.button_frame, bg=bg)
        self.slider_frame = tk.Frame(self.left_frame, bg=bg)
        self.control_buttons_frame = tk.Frame(self.left_frame, bg=bg)
        
        self.buttons['play'] = VWButton(self.control_buttons_frame, play_tk , _play)
        self.buttons['resume'] = VWButton(self.control_buttons_frame, play_tk, _resume)
        self.buttons['pause'] = VWButton(self.control_buttons_frame, tk.PhotoImage(file=BUTTON_PATH + buttons['pause']), _pause)
        self.buttons['stop'] = VWButton(self.control_buttons_frame, tk.PhotoImage(file=BUTTON_PATH + buttons['stop']), _stop)
        self.buttons['fast'] = VWButton(self.control_buttons_frame, tk.PhotoImage(file=BUTTON_PATH + buttons['fast']), _fast)
        self.buttons['reset'] = VWButton(self.control_buttons_frame, tk.PhotoImage(file=BUTTON_PATH + buttons['reset']), _reset)

        
        _img_dif = Image.open(BUTTON_PATH + buttons['difficulty'])
        self.buttons['difficulty'] = VWDifficultyButton(self.control_buttons_frame, _img_dif, _difficulty)
        
        self.pack_buttons('play', 'reset', 'fast', 'difficulty', forget=False)
        
        #init the slider
        self._init_size_slider(self.slider_frame, bg)
        
        self.slider_frame.grid(row=0,column=0)#.pack(side='top')
        self.control_buttons_frame.grid(row=1,column=0,sticky=tk.W)#.pack(side='bottom')

        self.left_frame.pack(side='left', fill=tk.X)
  
        #middle contains save and load
        self.mid_frame = tk.Frame(self.button_frame, bg=bg)
        self.saveload_frame = tk.Frame(self.mid_frame, bg=bg)
    
        #buttons
        self.buttons['save'] = VWButton(self.saveload_frame, tk.PhotoImage(file=BUTTON_PATH + buttons['save']), lambda: _save(self.load_menu))
        self.buttons['load'] = VWButton(self.saveload_frame, tk.PhotoImage(file=BUTTON_PATH + buttons['load']), lambda: _load(self.load_menu))
        #entry box
        files = saveload.files()
        self.load_menu = AutocompleteEntry(files, 4, self.mid_frame)
        self.load_menu.bind('<Button-1>', lambda _: self.deselect())
        self.load_menu.pack(side='top')
    
        self.pack_buttons('save', 'load', forget=False)
        self.saveload_frame.pack(side='bottom')
        self.mid_frame.pack(side='left')

        #init information frame
        self.info_frame = tk.Frame(self.button_frame, bg=bg)
        
        _size_frame = tk.Frame(self.info_frame, bg=bg)
        self.size_text = tk.StringVar()
        self.size_text.set(str(INITIAL_ENVIRONMENT_DIM))
        self.size_label = tk.Label(_size_frame, textvariable=self.size_text, width=2, font=ROOT_FONT, bg=bg)
        self.size_label.grid(row=0, column=1, sticky=tk.E)
        
        _size = tk.StringVar()
        _size.set("size:")
        _size_label =  tk.Label(_size_frame, textvariable=_size, font=ROOT_FONT, bg=bg)
        _size_label.grid(row=0, column=0, sticky=tk.W)
        _size_frame.grid(row=0, column=0, stick=tk.W)
        
        self.coordinate_text = tk.StringVar()
        self.coordinate_text.set("(-,-)")
        self.coordinate_label = tk.Label(self.info_frame, textvariable=self.coordinate_text, font=ROOT_FONT, bg=bg)
        self.coordinate_label.grid(row=1, column=0, sticky=tk.W)

        self.info_frame.pack(side='left', expand=True)
        self.button_frame.grid(row=1, column=0, pady=3, sticky=tk.W+tk.E)

        return buttons
    
    
    def _init_size_slider(self, parent, bg, length=200):
        #f = tk.Frame(parent, bg=bg)
        #t = tk.Label(f, text=" size ", bg=bg, font = ROOT_FONT)
        #t.pack(side='left')
        increments = Grid.GRID_MAX_SIZE - Grid.GRID_MIN_SIZE
        self.grid_scale_slider = Slider(parent, self.on_resize, self.on_resize_slide, None, length, 16,
                                        increments=increments,
                                        start=(DEFAULT_GRID_SIZE/DEFAULT_LOCATION_SIZE) - Grid.GRID_MIN_SIZE)
        self.grid_scale_slider.pack(side='top')
        #f.pack(side='bottom')
        
    def _init_dragables(self):
        #load all images
        keys = [('white', 'north'), ('orange', 'north'), ('green', 'north'), ('user', 'north'), ('orange', 'dirt'), ('green', 'dirt')]
        self.dragables = {}

        ix = DEFAULT_GRID_SIZE + DEFAULT_LOCATION_SIZE / 2 + 2
        iy = DEFAULT_LOCATION_SIZE / 2 + 4
        
        #print(self.all_images)
        for i, key in enumerate(keys):
            item = self.canvas.create_image(ix,
                                            iy + i * DEFAULT_LOCATION_SIZE, image=self.all_images_tk[key])
            drag_manager = CanvasDragManager(key, self.grid, self.canvas, item, self.drag_on_start, self.drag_on_drop)
            self.dragables[item] = (drag_manager, key)

    def deselect(self):
        self.selected = None
        if self.rectangle_selected:
            self.canvas.delete(self.rectangle_selected)
        self.rectangle_selected = None

    def select(self, event):
        if not self.running and _in_bounds(event.x, event.y):
            self.deselect()
            self.focus()
            inc = DEFAULT_GRID_SIZE / self.grid.dim
            coordinate = vwc.coord(int(event.x / inc), int(event.y / inc))
            print("SELECT:", self.grid.state[coordinate])
            self.selected = grid.state[coordinate]
            xx = coordinate.x * inc
            yy = coordinate.y * inc
            self.rectangle_selected = self.canvas.create_rectangle((xx, yy, xx + inc, yy + inc), fill='', width=3)


    def remove_top(self, event):
        if not self.running and _in_bounds(event.x, event.y):
            print("remove top")
            inc = DEFAULT_GRID_SIZE / self.grid.dim
            coordinate = vwc.coord(int(event.x / inc), int(event.y / inc))
            location = grid.state[coordinate]
            if location.agent:
                self.remove_agent(coordinate)
                self.grid.remove_agent(coordinate)
            elif grid.state[coordinate].dirt:
                self.remove_dirt(coordinate)
                self.grid.remove_dirt(coordinate)

    #remove an agent from the view
    def remove_dirt(self, coordinate):
        old = self.canvas_dirts[coordinate]
        self.canvas.delete(old)
        del old

    #remove an agent from the view
    def remove_agent(self, coordinate):
        old = self.canvas_agents[coordinate]
        self.canvas.delete(old)
        del old

    def rotate_agent(self, event, direction):
        #print('left', event)
        if self.selected and self.selected.agent:
            print(self.selected)
            self.remove_agent(self.selected.coordinate)
            new_orientation =  direction(self.selected.agent.orientation)
            inc = DEFAULT_GRID_SIZE / self.grid.dim
            tk_img = self.all_images_tk_scaled[(self.selected.agent.colour, new_orientation)]
            item = self.canvas.create_image(self.selected.coordinate.x * inc + inc/2,
                                            self.selected.coordinate.y * inc + inc/2, image=tk_img)
            self.canvas_agents[self.selected.coordinate] = item
            self.grid.turn_agent(self.selected.coordinate, new_orientation)
            self.selected = self.grid.state[self.selected.coordinate]
            self._lines_to_front()

    def rotate_agent_left(self, event):
        self.rotate_agent(event, vwc.left)

    def rotate_agent_right(self, event):
        self.rotate_agent(event, vwc.right)

    def pack_buttons(self, *buttons, forget=True):
        if forget:
            for button in self.buttons.values():
               button._button.grid_remove()
        for i in range(len(buttons)):
            self.buttons[buttons[i]]._button.grid(row=0, column=i, sticky=tk.W)

    def _reset_canvas(self, lines=True, dirts=True, agents=True, select=True):
        if lines:
            for line in self.grid_lines:
                self.canvas.delete(line)
            self.grid_lines.clear()
        if agents:
            for a in self.canvas_agents.values():
                self.canvas.delete(a)
            self.canvas_agents.clear()
        if dirts:
            for d in self.canvas_dirts.values():
                self.canvas.delete(d)
            self.canvas_dirts.clear()
        if select:
            self.deselect()

    def _lines_to_front(self):
        for line in self.grid_lines:
            self.canvas.tag_raise(line)
        if self.rectangle_selected:
            self.canvas.tag_raise(self.rectangle_selected)

    def _reset_canvas_agents(self):
        for a in self.canvas_agents.values():
            self.canvas.delete(a)
        self.canvas_agents.clear()

    def _redraw(self):
        self._reset_canvas(lines=False)
        inc = DEFAULT_GRID_SIZE / self.grid.dim
        for coord, location in grid.state.items():
            if location:
                if location.agent:
                    #print("AGENT:", coord, location)
                    tk_img = self.all_images_tk_scaled[(location.agent.colour, location.agent.orientation)]
                    item = self.canvas.create_image(coord.x * inc + inc/2,
                                                    coord.y * inc + inc/2, image=tk_img)
                    self.canvas_agents[coord] = item
                    self.canvas.tag_lower(item) #keep the agent behind the grid lines
                    if coord in self.canvas_dirts: #keep the dirt behind the agent
                        self.canvas.tag_lower(self.canvas_dirts[coord])
                if location.dirt:
                    #print("DDDIRT:",  coord, location)
                    tk_img = self.all_images_tk_scaled[(location.dirt.colour, 'dirt')]
                    item = self.canvas.create_image(coord.x * inc + inc/2, coord.y * inc + inc/2, image=tk_img)
                    self.canvas_dirts[coord] = item
                    self.canvas.tag_lower(item) #keep dirt behind agents and grid lines
        #self._lines_to_front()

    def _draw_grid(self, env_dim, size = DEFAULT_GRID_SIZE):
        x = 0
        y = 0
        inc = size / env_dim
        for i in range(env_dim + 1):
           self.grid_lines.append(self.canvas.create_line(x,0,x,size+1))
           self.grid_lines.append(self.canvas.create_line(0,y,size+1,y))
           y += inc
           x += inc

    def _get_image_key(self, name):
        s = name.split("_")
        return (s[0], s[1])

    def _init_images(self):
        # agents
        files = get_location_img_files(LOCATION_AGENT_IMAGES_PATH)
        image_names = [file.split('.')[0] for file in files]

        for img_name in image_names:
            file = os.path.join(LOCATION_AGENT_IMAGES_PATH, img_name) +  '.png'
            img = self._scale(Image.open(file), DEFAULT_LOCATION_SIZE)
            images = self._construct_images(img, img_name + '_')
            for img_name, img in images.items():
                img_key = self._get_image_key(img_name)
                tk_img = ImageTk.PhotoImage(img)
                self.all_images[img_key] = img
                self.all_images_tk[img_key] = tk_img

        # dirts
        files = get_location_img_files(LOCATION_DIRT_IMAGES_PATH)
        images_names = [file.split('.')[0] for file in files]

        for name in images_names:
            file = os.path.join(LOCATION_DIRT_IMAGES_PATH, name) +  '.png'
            img = self._scale(Image.open(file), DEFAULT_LOCATION_SIZE)
            img_key = self._get_image_key(name)
            tk_img = ImageTk.PhotoImage(img)
            self.all_images[img_key] = img
            self.all_images_tk[img_key] = tk_img

        self._scaled_tk()

    def _construct_images(self, img, name):
        #change this from magic strings... (use vwc orientation)
        return odict({name + 'north':img,
                      name + 'west':img.copy().rotate(90),
                      name + 'south':img.copy().rotate(180),
                      name + 'east':img.copy().rotate(270)})

    def _scaled_tk(self):
        size = min(DEFAULT_LOCATION_SIZE, DEFAULT_GRID_SIZE  / self.grid.dim)
        for name, image in self.all_images.items():
            self.all_images_tk_scaled[name] = ImageTk.PhotoImage(self._scale(image, size))

    def _scale(self, img, lsize):
        scale = lsize / max(img.width, img.height)
        return img.resize((int(img.width  * scale), int(img.height * scale)), Image.BICUBIC)

    #resize the grid
    def on_resize(self, value):
        value =  value + Grid.GRID_MIN_SIZE
        if value != self.grid.dim:
            self.grid.reset(value)
            self._reset_canvas()
            self._scaled_tk()
            self._draw_grid(grid.dim)
            
    def on_resize_slide(self, value):
        self.size_text.set(str(value + Grid.GRID_MIN_SIZE))
    
    def on_leave_canvas(self, event):
        self.coordinate_text.set('(-,-)')
    
    def on_mouse_move(self, event):
        if _in_bounds(event.x, event.y):
            inc = DEFAULT_GRID_SIZE / self.grid.dim
            self.coordinate_text.set('({},{})'.format(int(event.x / inc), int(event.y / inc)))
        else:
            self.coordinate_text.set('(-,-)')

    def drag_on_start(self, event):
        drag_manager, img_key = self.dragables[event.widget.find_closest(event.x, event.y)[0]]

        drag_manager.drag_image = self.all_images_tk_scaled[img_key]#ImageTk.PhotoImage(self._scale(image, size))

        drag_manager.drag = self.canvas.create_image(event.x, event.y, image=drag_manager.drag_image)

        self.canvas.itemconfigure(drag_manager.drag, state='hidden')
        self.canvas.tag_lower(drag_manager.drag)
        self.selected = None

        #keep the currently selected draggable on the top
        for a in self.canvas_agents.values():
            self.canvas.tag_lower(a)
        for d in self.canvas_dirts.values():
            self.canvas.tag_lower(d)

    def drag_on_drop(self, event, drag_manager):
        #TODO stream line to work with select

        inc = DEFAULT_GRID_SIZE / self.grid.dim
        x = int(event.x / inc)
        y = int(event.y / inc)
        coord = vwc.coord(x,y)
        #update the environment state
        colour, obj = drag_manager.key
        if obj == 'dirt':
            dirt1 =  grid.dirt(colour)
            grid.replace_dirt(coord, dirt1)
            if coord in self.canvas_dirts:
                self.canvas.delete(self.canvas_dirts[coord])
            self.canvas_dirts[coord] = drag_manager.drag
            self.canvas.tag_lower(self.canvas_dirts[coord])
        else: #its and agent
            agent1 =  grid.agent(colour, obj)
            grid.replace_agent(coord, agent1)
            if coord in self.canvas_agents:
                self.canvas.delete(self.canvas_agents[coord])
            self.canvas_agents[coord] = drag_manager.drag
        print("INFO: drop", self.grid.state[coord])
        
        self.select(event)

    def show_hide_side(self, state):
        for item in self.dragables.keys():
            self.canvas.itemconfigure(item, state=state)
        if state == 'hidden':
            self.grid_scale_slider.pack_forget()
            self.mid_frame.pack_forget()
            self.info_frame.pack_forget()
        elif state == 'normal':
            self.grid_scale_slider.pack(side='bottom')
            self.mid_frame.pack(side='left')
            self.info_frame.pack(side='left', expand=True)
        #self.canvas.itemconfig(self.options, state=state)

    def user_mind(self):
        return self.buttons['difficulty'].difficulty
    

def _fast():
    global TIME_STEP, TIME_STEP_BASE, TIME_STEP_MODIFIER
    TIME_STEP_MODIFIER /= 2.
    TIME_STEP = TIME_STEP_BASE * TIME_STEP_MODIFIER + TIME_STEP_MIN
    print("INFO: simulation speed set to {:1.4f} s/cycle".format(TIME_STEP))

def reset_time_step():
    global TIME_STEP, TIME_STEP_BASE, TIME_STEP_MODIFIER
    TIME_STEP_MODIFIER = 1.
    TIME_STEP = TIME_STEP_BASE * TIME_STEP_MODIFIER + TIME_STEP_MIN
    print("INFO: simulation speed set to {:1.4f} s/cycle".format(TIME_STEP))

def _difficulty():
    global user_mind
    user_mind = main_interface.user_mind()

def _save(saveloadmenu):
    file = saveloadmenu.var.get()
    if len(file) > 0:
        print('INFO: save', file)
        saveload.save(grid, file)
        saveloadmenu.lista = saveload.files()

def _load(saveloadmenu):
    file = saveloadmenu.var.get()
    if len(file) > 0:
        if not file.endswith('.vw'):
            file = file + ".vw"
        if file in saveloadmenu.lista:
            print('INFO: load:', file)
            data = saveload.load(file)
            if data is not None:
                main_interface.grid_scale_slider.set_position(data.dim - grid.GRID_MIN_SIZE)
                grid.replace_all(data)
                main_interface._redraw()

            
            
    
#resets the grid and enviroment
def _reset():
    print('INFO: reset')
    global reset
    reset = True
    main_interface._reset_canvas(lines=False)
    grid.reset(grid.dim)
    reset_time_step()

def _play():
    print('INFO: play')

    play_event.set()
    main_interface.pack_buttons('stop', 'pause', 'fast')
    main_interface.show_hide_side('hidden')
    main_interface.deselect()
    main_interface.running = True
              

def _stop():
    print('INFO: stop')
    global reset
    reset = True
    play_event.clear()
    reset_time_step()
    main_interface.running = False
    main_interface.pack_buttons('play', 'reset', 'fast', 'difficulty', 'save', 'load')
    main_interface.show_hide_side('normal')

def _resume():
    print('INFO: resume')
    
    play_event.set()
    main_interface.pack_buttons('stop', 'pause','fast')

def _pause():
    print('INFO: pause')
    play_event.clear()
    reset_time_step()
    main_interface.pack_buttons('stop', 'resume','fast')

def _back():
    print('INFO: back')
    play_event.clear()
    main_interface.pack_forget()
    main_menu.pack()

def _error(*_):
    traceback.print_exc()

def _finish():
    global root
    root.destroy()
    global finish
    finish = True
    
def _start():
    main_menu.pack_forget()
    main_interface.pack()

def _in_bounds(x,y):
    return x < DEFAULT_GRID_SIZE and x > 0 and y < DEFAULT_GRID_SIZE and y > 0

def run(_minds, skip = False, play = False, speed = 0, load = None):
    #required to avoid problems with running after errors (root must be destoryed!)
    assert(speed >= 0 and speed <= 1)
        
  #  import saveload
    global root
    tk.Tk.report_callback_exception = _error
    root = tk.Tk()
    root.title("Vacuum World")
    root.protocol("WM_DELETE_WINDOW", _finish)
    root.configure(background='white')
    
    try:
        global main_menu
        global main_interface
        global grid
        global minds
        minds = _minds

        global user_mind
        user_mind = 0

        grid = Grid(INITIAL_ENVIRONMENT_DIM)
        saveload.init()
        #saveload.load('/test.vw', grid)
        
        if not skip:    
            main_menu = VWMainMenu(root, _start, _finish)
            main_interface = VWInterface(root, grid)
        else:
            main_interface = VWInterface(root, grid)
            main_interface.pack()
            
        if play:
            if load is None:
                raise ValueError("argument \"load\" must be specified if argument play = True")
            load = saveload.format_file(load)
            files = saveload.files()
            if not load in files:
                raise ValueError("invalid file name:" + str(load) + "valid files include:" + str(files))
            print("INFO: autoplay enabled")
            
        if load is not None:
            grid.replace_all(saveload.load(load))
            main_interface._redraw()
            print("INFO: successfully loaded: ", load)
            
        global env_thread
        global play_event, finish, reset

        reset = True
        finish = False

        play_event = Event()

        env_thread = Thread(target=simulate, daemon=False)
        env_thread.start()
        
        #set up simulation speed
        global TIME_STEP
        global TIME_STEP_MODIFIER
        TIME_STEP_MODIFIER = 1 - speed
        TIME_STEP = TIME_STEP_BASE * TIME_STEP_MODIFIER + TIME_STEP_MIN
        print("INFO: simulation speed set to {:1.4f} s/cycle".format(TIME_STEP))
                
        if play:
            _play()
        root.mainloop()
        
    except:
        _error()
        _finish()




#TODO, be able to select user mind from gui

def simulate():
    try:
        def wait():
            should_sleep = not play_event.is_set()
            play_event.wait()
            if should_sleep:
                with Sleep(TIME_STEP*1000):
                     pass
            return True
        
        global reset
        global should_update
        global user_mind
        
        while wait() and not finish:
            #with TimeRecord(str(TIME_STEP)): #ms:
            with Sleep(TIME_STEP*1000):
                if reset:
                    global env
                    env = init_environment(grid, minds, user_mind)
                    grid.cycle = 0
                    reset = False
                
                print("------------ cycle {} ------------ ".format(grid.cycle))
                
                env.evolveEnvironment()
            
                grid.cycle += 1
                if not finish:
                    root.after(0, main_interface._redraw)
    except:
        print("INFO: SIMULATION ERROR")
        _error()
        root.after(0, _finish)
        
t = lambda: int(round(time.time() * 1000))

class Sleep:
    
    def __init__(self, wait):
        self.wait = wait
       
    def __enter__(self):
        self.start = t()
        self.finish = self.start + self.wait
    
    def __exit__(self, type, value, traceback):
        while t() < self.finish:
            time.sleep(1./1000.)  
            
class TimeRecord:
    
    def __init__(self, name):
        self._name = name
    
    def __enter__(self,):
        self._t = time.time() * 1000
    
    def __exit__(self, type, value, traceback):
        print("INFO: {0} time: {1}".format(self._name, time.time()  * 1000 - self._t))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul 28 12:00:27 2019

@author: ben
"""
from . import vwc
from .vwc import action



class EasyUser:
    
    def __init__(self):
        self.observation = None
        self.id = None
        self.move_actions = [action.move(), action.turn(vwc.direction.left), action.turn(vwc.direction.right)]
        self.actions = [action.drop(vwc.colour.green), action.drop(vwc.colour.orange)]
        self.actions.extend(self.move_actions)
        
    def decide(self): 

        if not self.observation.forward: #there is a wall infront
            if not self.observation.left:
                return action.turn(vwc.direction.right)
            if not self.observation.right:
                return action.turn(vwc.direction.left)
            return vwc.random(self.move_actions[1:])
        
        #if there is already a dirt at this location, move or turn
        if self.observation.center.dirt:
            if not self.observation.left:
                return vwc.random(self.move_actions, [0.6, 0.0, 0.4])
            if not self.observation.right:
                return vwc.random(self.move_actions, [0.6, 0.4, 0.0])
            return vwc.random(self.move_actions, [0.5, 0.25, 0.25])
    
        #otherwise do a random action (including dropping dirt)
        return vwc.random(self.actions, [0.2, 0.2, 0.45, 0.075, 0.075])
            
    def revise(self, observation, messages):
        self.id = observation.center.agent.name
        self.observation = observation
        
class MediumUser:

    def __init__(self):
        self.observation = None
        self.id = None
        
    def decide(self): 
        #there is a wall forward of the agent
        if not self.observation.forward:
            if not self.observation.left: #wall left and forward
                return action.turn(vwc.direction.right)
            if not self.observation.right: #wall right and forward
                return action.turn(vwc.direction.left)
            if self.observation.left.agent and self.observation.right.agent: #both left and right are full
                return vwc.random([action.turn(vwc.direction.left), action.turn(vwc.direction.right)])
            if self.observation.left.agent: #the left location is full
                return action.turn(vwc.direction.right)
            if self.observation.right.agent: #the right location is full
                return action.turn(vwc.direction.left)
            if self.observation.center.dirt: #both left and right are free, drop dirt?
                return vwc.random([action.turn(vwc.direction.left), action.turn(vwc.direction.right)])
            return vwc.random([action.turn(vwc.direction.left), action.turn(vwc.direction.right),
                               action.drop(vwc.colour.green), action.drop(vwc.colour.orange)])
            
        if self.observation.forward.agent:
            if not self.observation.left:
                return action.turn(vwc.direction.right)
            if not self.observation.right:
                return action.turn(vwc.direction.left)
            if self.observation.left.agent and self.observation.right.agent: #both left and right are full
                return vwc.random([action.drop(vwc.colour.green), action.drop(vwc.colour.orange)])
            if self.observation.left.agent: #the left location is full
                return action.turn(vwc.direction.right)
            if self.observation.right.agent: #the right location is full
                return action.turn(vwc.direction.left)
            if self.observation.center.dirt: #both left and right are free, drop dirt?
                return vwc.random([action.turn(vwc.direction.left), action.turn(vwc.direction.right)])
            return vwc.random([action.turn(vwc.direction.left), action.turn(vwc.direction.right),
                               action.drop(vwc.colour.green), action.drop(vwc.colour.orange)])
        
        #if there is an agent in some direction, turn to face away from it aslong as there isnt a wall
        if self.observation.left and self.observation.left.agent and self.observation.right and self.observation.right.agent:
            return self.move_or_drop()
        if self.observation.left and self.observation.left.agent and self.observation.right:
            return vwc.random([action.turn(vwc.direction.right), action.move()])
        if self.observation.right and self.observation.right.agent and self.observation.left:
            return vwc.random([action.turn(vwc.direction.left), action.move()])
        
        if not self.observation.left:
            if self.observation.center.dirt:
                return vwc.random([action.move(), action.turn(vwc.direction.right)], [0.9, 0.1])
            return vwc.random([action.move(), action.turn(vwc.direction.right),
                               action.drop(vwc.colour.green), action.drop(vwc.colour.orange)], [0.6, 0.25, 0.075, 0.075])
        
        if not self.observation.right:
            if self.observation.center.dirt:
                return vwc.random([action.move(), action.turn(vwc.direction.left)], [0.9, 0.1])
            return vwc.random([action.move(), action.turn(vwc.direction.left),
                               action.drop(vwc.colour.green), action.drop(vwc.colour.orange)], [0.6, 0.25, 0.075, 0.075])
        return self.random_all()         
    
    def move_or_drop(self):
        return vwc.random([action.move(), action.drop(vwc.colour.green), action.drop(vwc.colour.orange)], [0.8, 0.1, 0.1])   
     
    def random_all(self):
        return vwc.random([action.move(), action.drop(vwc.colour.green), action.drop(vwc.colour.orange),
                           action.turn(vwc.direction.left), action.turn(vwc.direction.right)], [0.6, 0.15, 0.15, 0.05, 0.05]) 
            
    def revise(self, observation, messages):
        self.id = observation.center.agent.name
        self.observation = observation

#the users that can be used in vacuumworld
USERS = [EasyUser, MediumUser]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Sep 25 10:29:55 2019

@author: ben
"""
from pystarworlds.Agent import Mind

import inspect

def callerID():
    caller = inspect.currentframe().f_back
    while not isinstance(caller.f_locals.get('self', None), Mind):
        caller = caller.f_back
    return caller.f_locals['self'].body.ID

def warn_agent(message, *args):
    print(message.format(callerID(), *args))


from pystarworlds.Agent import Actuator

from . import vwaction

class CleaningAgentActuator(Actuator):
    subscribe = [vwaction.MoveAction, vwaction.TurnAction, vwaction.CleanAction]

class UserActuator(Actuator):
    subscribe = [vwaction.MoveAction,vwaction.TurnAction,vwaction.DropAction]
    
class CommunicationActuator(Actuator):
    subscribe = [vwaction.CommunicativeAction]
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jun  2 21:22:31 2019

@author: ben
"""

import setuptools

setuptools.setup(name='vacuumworld',
      version='4.1.2',
      description='',
      url='https://github.com/dicelab-rhul/vacuumworld',
      author='Benedict Wilkins',
      author_email='brjw@hotmail.co.uk',
      license='GNU General Public License v3 (GPLv3)T',
      packages=['vacuumworld'],
      include_package_data=True,
      install_requires=['pystarworlds', 'pillow'],
      classifiers=[
        "Programming Language :: Python :: 3.7",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
      ])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jun  1 21:53:28 2019

@author: ben
"""
import random as rand
from collections import namedtuple

location = namedtuple('location', 'coordinate agent dirt')
agent = namedtuple('agent', 'name colour orientation')
dirt = namedtuple('dirt', 'name colour')
_coord = namedtuple('coordinate', 'x y')

class coord(_coord):
    
    def __add__(self, other):
        return coord(self[0] + other[0], self[1] + other[1])
    
    def __sub__(self, other):
        return coord(self[0] - other[0], self[1] - other[1])
    
    def __mul__(self, other):
        return coord(self[0] * other[0], self[1] * other[1])
    
    def __div__(self, other):
        return coord(int(self[0] / other[0]), int(self[1] / other[1]))
    


perception = namedtuple('perception', 'observation messages')

message = namedtuple('message', 'sender content')

_observation = namedtuple('observation', 'center left right forward forwardleft forwardright')

class observation(_observation):
    def __iter__(self):
        return (x for x in super(observation, self).__iter__() if x is not None)

orientation = namedtuple('orientation', 'north east south west')('north', 'east', 'south', 'west')

orientation_map = {orientation.north:coord(0,-1), 
                   orientation.south:coord(0,1), 
                   orientation.west:coord(-1,0), 
                   orientation.east:coord(1,0)}

def left(_orientation):
    return orientation[(orientation.index(_orientation) - 1) % 4]

def right(_orientation):
    return orientation[(orientation.index(_orientation) + 1) % 4]

direction = namedtuple('direction', 'left right')(left, right)

colour = namedtuple('colour', 'white orange green user')('white', 'orange', 'green', 'user')

_colour_dirt = set(['orange','green'])
_colour_agent = set(['orange','green','white'])
_colour_user = set(['user'])

move = namedtuple('move', '')
turn = namedtuple('turn', 'direction')
clean = namedtuple('clean', '')
drop = namedtuple('drop', 'colour')
idle = namedtuple('idle', '')
speak = namedtuple('speak', 'message to')
action_types = namedtuple('actions_types', 'move turn clean drop idle speak')(move, turn, clean, drop, idle, speak)

action = namedtuple('actions', 'move turn clean drop idle speak')(
                    lambda : move(), 
                    lambda _direction : turn(_direction),  
                    lambda : clean(),
                    lambda _colour: drop(_colour),
                    lambda : idle(),
                    lambda message, *to : speak(message, to))

def size(message):
    _size = 0
    if type(message) in (list, tuple): 
        _size += len(message) + 1
        for e in message:
            _size += size(e)
    elif message is not None:
        _size += len(str(message))
    return _size

def random(actions, p=None):
    if p is None:
        return rand.choice(actions)
    else:
        assert len(actions) == len(p)
        return rand.choices(actions, weights=p, k=1)[0]



